import { rollup, watch as watch$1 } from 'rollup';
import dts from 'rollup-plugin-dts';
import merge from 'lodash.merge';
import { resolve, dirname, basename, join } from 'pathe';
import { readFile, rm } from 'fs/promises';
import alias from '@rollup/plugin-alias';
import commonjs from '@rollup/plugin-commonjs';
import json from '@rollup/plugin-json';
import resolve$1 from '@rollup/plugin-node-resolve';
import replace from '@rollup/plugin-replace';
import run from '@rollup/plugin-run';
import RollupEsbuild from 'rollup-plugin-esbuild';
import { createFilter } from '@rollup/pluginutils';
import { transform } from '@swc/core';
import { existsSync } from 'fs';
import RollupTypescript from 'rollup-plugin-typescript2';
import jiti from 'jiti';
import Module from 'module';
import mri from 'mri';

const defineOutput = (output) => output;

let packageJson = {};
async function loadPackage(rootDir, force) {
  if (packageJson.name && !force) {
    return Promise.resolve(packageJson);
  }
  const raw = await readFile(resolve(rootDir, "package.json"));
  packageJson = JSON.parse(raw.toString());
  return Promise.resolve(packageJson);
}

async function loadTsConfig(rootDir, filename) {
  const resolved = resolve(rootDir, filename);
  const readed = await readFile(resolved, { encoding: "utf-8" });
  const tsConfig = JSON.parse(readed);
  tsConfig.compilerOptions = tsConfig.compilerOptions || {};
  if (typeof tsConfig.extends === "string") {
    const filepath = resolve(dirname(resolved), tsConfig.extends);
    const parent = await loadTsConfig(dirname(filepath), basename(filepath));
    tsConfig.compilerOptions = merge(parent.compilerOptions || {}, tsConfig.compilerOptions);
  }
  return tsConfig;
}

function remove(path) {
  return rm(path, { force: true, recursive: true });
}

const outputMapping = /* @__PURE__ */ new Map();
function getRollupOutput(id, preserveModules = false) {
  const _output = outputMapping.get(id);
  return merge(_output, { preserveModules });
}
function getRollupOutputs(options) {
  _registerDefaultOutputs(options.outDir);
  return options.outputs.map((out) => _resolveOutput(options, out));
}
function registerOutput(id, output, force) {
  if (!outputMapping.has(id) || force) {
    outputMapping.set(id, output);
  }
}
function _registerDefaultOutputs(outDir) {
  const jsOutput = defineOutput({
    dir: outDir,
    entryFileNames: "[name].js",
    chunkFileNames: "chunks/[hash].js",
    format: "cjs",
    exports: "auto",
    preferConst: true,
    externalLiveBindings: false,
    freeze: false
  });
  const cjsOutput = defineOutput({
    dir: outDir,
    entryFileNames: "[name].cjs",
    chunkFileNames: "chunks/[hash].cjs",
    format: "cjs",
    exports: "auto",
    preferConst: true,
    externalLiveBindings: false,
    freeze: false
  });
  const esmOutput = defineOutput({
    dir: outDir,
    entryFileNames: "[name].mjs",
    chunkFileNames: "chunks/[hash].mjs",
    format: "esm",
    exports: "auto",
    preferConst: true,
    externalLiveBindings: false,
    freeze: false
  });
  const dtsOutput = defineOutput({
    dir: outDir,
    format: "esm"
  });
  registerOutput("js", jsOutput);
  registerOutput("cjs", cjsOutput);
  registerOutput("esm", esmOutput);
  registerOutput("__dts__", dtsOutput);
}
function _resolveOutput(options, output) {
  if (typeof output !== "string") {
    const outDir = resolve(options.rootDir, output.dir || "");
    const assigns = { dir: outDir };
    return merge(output, assigns);
  }
  if (!outputMapping.has(output)) {
    console.error(`${output} output type is not registered`);
    process.exit(1);
  }
  return getRollupOutput(output, options.preserveModules);
}

async function createBuilder(options, rollupOptions) {
  const parallelBuilds = [];
  const allOutputOptions = rollupOptions.output;
  const buildResult = await rollup(rollupOptions);
  console.info(`Building for ${options.target}`);
  for (const outputOptions of allOutputOptions) {
    const writter = buildResult.write(outputOptions);
    parallelBuilds.push(writter);
  }
  if (options.declaration) {
    rollupOptions.plugins?.push(dts({}));
    const typesBuild = await rollup(rollupOptions);
    const typesOutput = getRollupOutput("__dts__", options.preserveModules);
    const typesWritter = typesBuild.write(typesOutput);
    parallelBuilds.push(typesWritter);
  }
  return Promise.all(parallelBuilds);
}

function createWatcher(options, rollupOptions) {
  if (Array.isArray(rollupOptions.output)) {
    rollupOptions.output = rollupOptions.output.map((out) => _transformOutput(out));
  }
  console.info("Watching for file changes...");
  const watcher = watch$1(rollupOptions);
  watcher.on("event", _createOnEventHandler(options));
  watcher.close();
  return watcher;
}
function _createOnEventHandler(options) {
  return async (event) => {
    if (event.code === "BUNDLE_START") {
      await options.onBundleStart?.();
      console.info("Build started...");
    } else if (event.code === "BUNDLE_END") {
      event.result.close();
      await options.onBundleEnd?.();
      console.info(`Built in ${event.duration}ms.`);
    } else if (event.code === "ERROR") {
      console.error(event.error.message);
    }
  };
}
function _transformOutput(outputOptions) {
  const chunkFileNames = outputOptions.chunkFileNames;
  if (typeof chunkFileNames === "string") {
    outputOptions.chunkFileNames = chunkFileNames.replace("[hash]", "[name]");
  }
  return outputOptions;
}

function esbuild(options, tsConfig) {
  const esbuildOptions = _getEsbuildOptions(options, tsConfig);
  return RollupEsbuild(esbuildOptions);
}
function _getEsbuildOptions(options, tsConfig) {
  const esbuildOptions = options.rollupPlugins?.esbuild || {};
  const compilerOptions = tsConfig?.compilerOptions;
  const jsx = compilerOptions?.jsx || "preserve";
  const defaultOptions = {
    include: esbuildOptions.include || /\.[jt]sx?$/,
    exclude: /node_modules/,
    sourceMap: compilerOptions?.sourceMap,
    minify: esbuildOptions.minify ?? process.env.NODE_ENV === "production",
    target: compilerOptions?.target,
    jsx: jsx === "preserve" ? "preserve" : "transform",
    jsxFactory: compilerOptions?.jsxFactory,
    jsxFragment: compilerOptions?.jsxFragmentFactory,
    tsconfig: esbuildOptions.tsconfig || options.tsConfig
  };
  return merge(defaultOptions, esbuildOptions);
}

function raw(options) {
  const filter = createFilter(options.include, options.exclude);
  return {
    name: "raw",
    async transform(code, id) {
      if (!filter(id)) {
        return null;
      }
      return {
        code: `export default ${JSON.stringify(code)}`,
        map: null
      };
    }
  };
}

function swc(options, tsConfig) {
  const swcOptions = options.rollupPlugins?.swc || {};
  const filter = createFilter(/\.[jt]sx?$/, /node_modules/);
  const pathResolver = _createPathResolver(options.rootDir, swcOptions.extensions);
  return {
    name: "swc",
    async resolveId(source, importer) {
      let resolvedPath = pathResolver(source, importer);
      if (!resolvedPath) {
        resolvedPath = pathResolver(source, importer, true);
      }
      return resolvedPath;
    },
    async transform(code, id) {
      if (!filter(id)) {
        return null;
      }
      const _swcOptions = _getSwcOptions(id, tsConfig, swcOptions);
      const transformed = await transform(code, _swcOptions);
      return {
        code: transformed.code,
        map: transformed.map && JSON.parse(transformed.map)
      };
    },
    async renderChunk(code, chunk) {
      if (!swcOptions.minify) {
        return null;
      }
      const transformed = await transform(code, {
        minify: true,
        sourceMaps: true,
        filename: chunk.fileName
      });
      return {
        code: transformed.code,
        map: transformed.map
      };
    }
  };
}
function _createPathResolver(rootDir, extensions = []) {
  return (source, importer, isIndex) => {
    let resolvedPath;
    const basedir = importer ? dirname(importer) : rootDir;
    const filepath = resolve(basedir, source);
    for (const extension of extensions) {
      const _filepath = isIndex ? join(filepath, `index${extension}`) : join(`${extension}${extension}`);
      if (existsSync(_filepath)) {
        resolvedPath = _filepath;
        break;
      }
    }
    return resolvedPath;
  };
}
function _getSwcOptions(id, tsConfig, options) {
  const isTsFile = /\.tsx?$/.test(id);
  const compilerOptions = tsConfig?.compilerOptions;
  const moduleTarget = compilerOptions?.target?.toString().toLowerCase();
  const defaultOptions = {
    filename: id,
    sourceMaps: compilerOptions?.sourceMap,
    jsc: {
      keepClassNames: compilerOptions?.experimentalDecorators,
      target: moduleTarget,
      parser: {
        syntax: isTsFile ? "typescript" : "ecmascript",
        tsx: isTsFile,
        jsx: !isTsFile,
        decorators: compilerOptions?.experimentalDecorators
      },
      transform: {
        legacyDecorator: compilerOptions?.experimentalDecorators,
        decoratorMetadata: compilerOptions?.emitDecoratorMetadata,
        react: {
          pragma: compilerOptions?.jsxFactory,
          pragmaFrag: compilerOptions?.jsxFragmentFactory,
          importSource: compilerOptions?.jsxImportSource
        }
      }
    }
  };
  return merge(defaultOptions, options);
}

function typescript(options, tsConfig) {
  const typescriptOptions = _getTypescriptOptions(options, tsConfig);
  return RollupTypescript(typescriptOptions);
}
function _getTypescriptOptions(options, tsConfig) {
  const typescriptOptions = options.rollupPlugins?.typescript || {};
  const compilerOptions = tsConfig?.compilerOptions;
  const defaultOptions = {
    tsconfig: "tsconfig.json",
    tsconfigOverride: { compilerOptions, declaration: false },
    useTsconfigDeclarationDir: true
  };
  return merge(defaultOptions, typescriptOptions);
}

async function getRollupPlugins(options, watch) {
  const defaultPlugins = await getDefaultPlugins(options, watch);
  return typeof options.mapRollupPlugins === "function" ? options.mapRollupPlugins(defaultPlugins) : defaultPlugins;
}
async function getDefaultPlugins(options, watch) {
  const tsConfigJson = await loadTsConfig(options.rootDir, options.tsConfig);
  const syntaxParser = resolveSyntaxParser(options, tsConfigJson);
  const runEnabled = watch && options.run;
  return [
    replace({
      true: true,
      ...options.rollupPlugins?.replace
    }),
    alias({
      ...options.rollupPlugins?.alias
    }),
    resolve$1({
      mainFields: ["module", "jsnext", "main"],
      browser: options.target === "browser",
      extensions: options.extensions,
      exportConditions: [options.target === "node" ? "node" : "browser"],
      preferBuiltins: options.target === "node",
      ...options.rollupPlugins?.resolve
    }),
    json({
      preferConst: true,
      ...options.rollupPlugins?.json
    }),
    syntaxParser(),
    commonjs({
      include: [/node_modules/],
      extensions: options.extensions,
      ignoreTryCatch: true,
      requireReturnsDefault: "namespace",
      ...options.rollupPlugins?.commonjs
    }),
    raw({
      include: [/\.(md|txt)$/],
      exclude: [],
      ...options.rollupPlugins?.raw
    }),
    runEnabled && run({
      ...options.rollupPlugins?.run
    })
  ];
}
function resolveSyntaxParser(options, tsConfigJson) {
  const factory = _resolveParserFactory(options.parser);
  return () => factory(options, tsConfigJson);
}
function _resolveParserFactory(parser) {
  switch (parser) {
    case "swc":
      return swc;
    case "typescript":
      return typescript;
    default:
      return esbuild;
  }
}

async function getRollupOptions(options, watch) {
  const rollupPlugins = await getRollupPlugins(options, watch);
  const defaultOptions = {
    context: options.rootDir,
    input: options.entries,
    output: getRollupOutputs(options),
    plugins: rollupPlugins,
    watch: watch && _getWatchOptions(),
    external: _createExternalFilter(options),
    onwarn: _createWarnHandler()
  };
  return merge(defaultOptions, options.rollupOptions);
}
function _getWatchOptions() {
  return {
    chokidar: {
      ignoreInitial: true,
      ignorePermissionErrors: true,
      ignored: ["**/.git/**", "**/node_modules/**"]
    }
  };
}
function _getPackageName(id) {
  const partials = id.split("/");
  return partials[0][0] === "@" ? `${partials[0]}/${partials[1]}` : partials[0];
}
function _createExternalFilter(options) {
  return (id) => {
    const packageName = _getPackageName(id);
    const isExternal = options.externals.includes(packageName);
    if (isExternal) {
      return true;
    }
    if (options.inlineDependencies || id.startsWith(".") || id.startsWith("/") || id.match(/src[\\/]/)) {
      return false;
    }
    return isExternal;
  };
}
function _createWarnHandler() {
  const warningIgnoreList = ["CIRCULAR_DEPENDENCY", "EMPTY_BUNDLE", "THIS_IS_UNDEFINED"];
  return (warning, handler) => {
    if (warning.code === "UNRESOLVED_IMPORT") {
      const id = warning.source;
      const importer = warning.importer;
      if (!importer || !/\?commonjs-external$/.test(importer)) {
        console.error(`Rollup failed to resolve import "${id}" from "${importer}"`);
        process.exit(-1);
      }
    }
    if (!warningIgnoreList.includes(warning.code)) {
      if (warning.code === "PLUGIN_WARNING") {
        console.warn(`[plugin:${warning.plugin}] ${warning.message}`);
      } else {
        handler(warning);
      }
    }
  };
}

async function build(options) {
  const rollupOptions = await getRollupOptions(options);
  await remove(options.outDir);
  return createBuilder(options, rollupOptions);
}

async function watch(options) {
  const rollupOptions = await getRollupOptions(options, true);
  await remove(options.outDir);
  return createWatcher(options, rollupOptions);
}

const configNames = /* @__PURE__ */ new Set([
  ".bauenrc.js",
  ".bauenrc.cjs",
  ".bauenrc.mjs",
  ".bauenrc.ts",
  "bauen.config.js",
  "bauen.config.cjs",
  "bauen.config.mjs",
  "bauen.config.ts"
]);
function searchConfig(rootDir) {
  const result = { name: null, path: null };
  for (const name of configNames) {
    const path = resolve(rootDir, name);
    if (existsSync(path)) {
      Object.assign(result, { name, path });
      break;
    }
  }
  return result;
}

function loadConfig(rootDir) {
  const _require = jiti(rootDir, { interopDefault: true });
  const foundConfig = searchConfig(rootDir);
  try {
    return _require(foundConfig.path);
  } catch (err) {
    console.error(`Config file not found in ${rootDir}`, err);
    process.exit(1);
  }
}

function getCliOptions(argv) {
  const args = mri([...argv].splice(2));
  const { _, ...restArgs } = args;
  return {
    ..._.length > 0 && { command: _[0] },
    ...restArgs
  };
}

function getDefaultOptions() {
  return {
    command: "build",
    rootDir: process.cwd(),
    outDir: "dist",
    entries: [],
    outputs: [],
    target: "node",
    extensions: [".ts", ".tsx", ".mjs", ".cjs", ".js", ".jsx"],
    externals: [],
    tsConfig: "tsconfig.json",
    parser: "esbuild",
    run: false,
    declaration: true,
    inlineDependencies: false
  };
}

async function getOptions(argv, userConfig) {
  const defaultOptions = getDefaultOptions();
  const inputOptions = { ...getCliOptions(argv), ...userConfig };
  const rootDir = resolve(inputOptions.rootDir || defaultOptions.rootDir);
  const tsConfig = resolve(rootDir, inputOptions.tsConfig || defaultOptions.tsConfig);
  const [packageJson] = await Promise.all([
    loadPackage(rootDir),
    loadTsConfig(rootDir, tsConfig)
  ]);
  return merge(defaultOptions, inputOptions, {
    rootDir,
    outDir: resolve(rootDir, inputOptions.outDir || defaultOptions.outDir),
    entries: inputOptions.entries?.map((entry) => resolve(rootDir, entry)),
    externals: _getExternals(userConfig, packageJson)
  });
}
function _getExternals(userConfig, packageJson) {
  if (userConfig.target === "browser") {
    return [];
  }
  return [
    ...Module.builtinModules,
    ...Module.builtinModules.map((mod) => `node:${mod}`),
    ...Object.keys(packageJson.dependencies || []),
    ...Object.keys(packageJson.peerDependencies || []),
    ...userConfig.externals || []
  ];
}

export { createWatcher as a, build as b, createBuilder as c, getRollupOptions as d, getRollupOutput as e, getRollupOutputs as f, getOptions as g, getRollupPlugins as h, getCliOptions as i, getDefaultOptions as j, esbuild as k, loadConfig as l, raw as m, swc as n, defineOutput as o, loadPackage as p, loadTsConfig as q, registerOutput as r, searchConfig as s, typescript as t, remove as u, watch as w };
