'use strict';

const rollup = require('rollup');
const dts = require('rollup-plugin-dts');
const merge = require('lodash.merge');
const pathe = require('pathe');
const promises = require('fs/promises');
const alias = require('@rollup/plugin-alias');
const commonjs = require('@rollup/plugin-commonjs');
const json = require('@rollup/plugin-json');
const resolve = require('@rollup/plugin-node-resolve');
const replace = require('@rollup/plugin-replace');
const run = require('@rollup/plugin-run');
const RollupEsbuild = require('rollup-plugin-esbuild');
const pluginutils = require('@rollup/pluginutils');
const core = require('@swc/core');
const fs = require('fs');
const RollupTypescript = require('rollup-plugin-typescript2');
const jiti = require('jiti');
const Module = require('module');
const mri = require('mri');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e["default"] : e; }

const dts__default = /*#__PURE__*/_interopDefaultLegacy(dts);
const merge__default = /*#__PURE__*/_interopDefaultLegacy(merge);
const alias__default = /*#__PURE__*/_interopDefaultLegacy(alias);
const commonjs__default = /*#__PURE__*/_interopDefaultLegacy(commonjs);
const json__default = /*#__PURE__*/_interopDefaultLegacy(json);
const resolve__default = /*#__PURE__*/_interopDefaultLegacy(resolve);
const replace__default = /*#__PURE__*/_interopDefaultLegacy(replace);
const run__default = /*#__PURE__*/_interopDefaultLegacy(run);
const RollupEsbuild__default = /*#__PURE__*/_interopDefaultLegacy(RollupEsbuild);
const RollupTypescript__default = /*#__PURE__*/_interopDefaultLegacy(RollupTypescript);
const jiti__default = /*#__PURE__*/_interopDefaultLegacy(jiti);
const Module__default = /*#__PURE__*/_interopDefaultLegacy(Module);
const mri__default = /*#__PURE__*/_interopDefaultLegacy(mri);

const defineOutput = (output) => output;

let packageJson = {};
async function loadPackage(rootDir, force) {
  if (packageJson.name && !force) {
    return Promise.resolve(packageJson);
  }
  const raw = await promises.readFile(pathe.resolve(rootDir, "package.json"));
  packageJson = JSON.parse(raw.toString());
  return Promise.resolve(packageJson);
}

async function loadTsConfig(rootDir, filename) {
  const resolved = pathe.resolve(rootDir, filename);
  const readed = await promises.readFile(resolved, { encoding: "utf-8" });
  const tsConfig = JSON.parse(readed);
  tsConfig.compilerOptions = tsConfig.compilerOptions || {};
  if (typeof tsConfig.extends === "string") {
    const filepath = pathe.resolve(pathe.dirname(resolved), tsConfig.extends);
    const parent = await loadTsConfig(pathe.dirname(filepath), pathe.basename(filepath));
    tsConfig.compilerOptions = merge__default(parent.compilerOptions || {}, tsConfig.compilerOptions);
  }
  return tsConfig;
}

function remove(path) {
  return promises.rm(path, { force: true, recursive: true });
}

const outputMapping = /* @__PURE__ */ new Map();
function getRollupOutput(id, preserveModules = false) {
  const _output = outputMapping.get(id);
  return merge__default(_output, { preserveModules });
}
function getRollupOutputs(options) {
  _registerDefaultOutputs(options.outDir);
  return options.outputs.map((out) => _resolveOutput(options, out));
}
function registerOutput(id, output, force) {
  if (!outputMapping.has(id) || force) {
    outputMapping.set(id, output);
  }
}
function _registerDefaultOutputs(outDir) {
  const jsOutput = defineOutput({
    dir: outDir,
    entryFileNames: "[name].js",
    chunkFileNames: "chunks/[hash].js",
    format: "cjs",
    exports: "auto",
    preferConst: true,
    externalLiveBindings: false,
    freeze: false
  });
  const cjsOutput = defineOutput({
    dir: outDir,
    entryFileNames: "[name].cjs",
    chunkFileNames: "chunks/[hash].cjs",
    format: "cjs",
    exports: "auto",
    preferConst: true,
    externalLiveBindings: false,
    freeze: false
  });
  const esmOutput = defineOutput({
    dir: outDir,
    entryFileNames: "[name].mjs",
    chunkFileNames: "chunks/[hash].mjs",
    format: "esm",
    exports: "auto",
    preferConst: true,
    externalLiveBindings: false,
    freeze: false
  });
  const dtsOutput = defineOutput({
    dir: outDir,
    format: "esm"
  });
  registerOutput("js", jsOutput);
  registerOutput("cjs", cjsOutput);
  registerOutput("esm", esmOutput);
  registerOutput("__dts__", dtsOutput);
}
function _resolveOutput(options, output) {
  if (typeof output !== "string") {
    const outDir = pathe.resolve(options.rootDir, output.dir || "");
    const assigns = { dir: outDir };
    return merge__default(output, assigns);
  }
  if (!outputMapping.has(output)) {
    console.error(`${output} output type is not registered`);
    process.exit(1);
  }
  return getRollupOutput(output, options.preserveModules);
}

async function createBuilder(options, rollupOptions) {
  const parallelBuilds = [];
  const allOutputOptions = rollupOptions.output;
  const buildResult = await rollup.rollup(rollupOptions);
  console.info(`Building for ${options.target}`);
  for (const outputOptions of allOutputOptions) {
    const writter = buildResult.write(outputOptions);
    parallelBuilds.push(writter);
  }
  if (options.declaration) {
    rollupOptions.plugins?.push(dts__default({}));
    const typesBuild = await rollup.rollup(rollupOptions);
    const typesOutput = getRollupOutput("__dts__", options.preserveModules);
    const typesWritter = typesBuild.write(typesOutput);
    parallelBuilds.push(typesWritter);
  }
  return Promise.all(parallelBuilds);
}

function createWatcher(options, rollupOptions) {
  if (Array.isArray(rollupOptions.output)) {
    rollupOptions.output = rollupOptions.output.map((out) => _transformOutput(out));
  }
  console.info("Watching for file changes...");
  const watcher = rollup.watch(rollupOptions);
  watcher.on("event", _createOnEventHandler(options));
  watcher.close();
  return watcher;
}
function _createOnEventHandler(options) {
  return async (event) => {
    if (event.code === "BUNDLE_START") {
      await options.onBundleStart?.();
      console.info("Build started...");
    } else if (event.code === "BUNDLE_END") {
      event.result.close();
      await options.onBundleEnd?.();
      console.info(`Built in ${event.duration}ms.`);
    } else if (event.code === "ERROR") {
      console.error(event.error.message);
    }
  };
}
function _transformOutput(outputOptions) {
  const chunkFileNames = outputOptions.chunkFileNames;
  if (typeof chunkFileNames === "string") {
    outputOptions.chunkFileNames = chunkFileNames.replace("[hash]", "[name]");
  }
  return outputOptions;
}

function esbuild(options, tsConfig) {
  const esbuildOptions = _getEsbuildOptions(options, tsConfig);
  return RollupEsbuild__default(esbuildOptions);
}
function _getEsbuildOptions(options, tsConfig) {
  const esbuildOptions = options.rollupPlugins?.esbuild || {};
  const compilerOptions = tsConfig?.compilerOptions;
  const jsx = compilerOptions?.jsx || "preserve";
  const defaultOptions = {
    include: esbuildOptions.include || /\.[jt]sx?$/,
    exclude: /node_modules/,
    sourceMap: compilerOptions?.sourceMap,
    minify: esbuildOptions.minify ?? process.env.NODE_ENV === "production",
    target: compilerOptions?.target,
    jsx: jsx === "preserve" ? "preserve" : "transform",
    jsxFactory: compilerOptions?.jsxFactory,
    jsxFragment: compilerOptions?.jsxFragmentFactory,
    tsconfig: esbuildOptions.tsconfig || options.tsConfig
  };
  return merge__default(defaultOptions, esbuildOptions);
}

function raw(options) {
  const filter = pluginutils.createFilter(options.include, options.exclude);
  return {
    name: "raw",
    async transform(code, id) {
      if (!filter(id)) {
        return null;
      }
      return {
        code: `export default ${JSON.stringify(code)}`,
        map: null
      };
    }
  };
}

function swc(options, tsConfig) {
  const swcOptions = options.rollupPlugins?.swc || {};
  const filter = pluginutils.createFilter(/\.[jt]sx?$/, /node_modules/);
  const pathResolver = _createPathResolver(options.rootDir, swcOptions.extensions);
  return {
    name: "swc",
    async resolveId(source, importer) {
      let resolvedPath = pathResolver(source, importer);
      if (!resolvedPath) {
        resolvedPath = pathResolver(source, importer, true);
      }
      return resolvedPath;
    },
    async transform(code, id) {
      if (!filter(id)) {
        return null;
      }
      const _swcOptions = _getSwcOptions(id, tsConfig, swcOptions);
      const transformed = await core.transform(code, _swcOptions);
      return {
        code: transformed.code,
        map: transformed.map && JSON.parse(transformed.map)
      };
    },
    async renderChunk(code, chunk) {
      if (!swcOptions.minify) {
        return null;
      }
      const transformed = await core.transform(code, {
        minify: true,
        sourceMaps: true,
        filename: chunk.fileName
      });
      return {
        code: transformed.code,
        map: transformed.map
      };
    }
  };
}
function _createPathResolver(rootDir, extensions = []) {
  return (source, importer, isIndex) => {
    let resolvedPath;
    const basedir = importer ? pathe.dirname(importer) : rootDir;
    const filepath = pathe.resolve(basedir, source);
    for (const extension of extensions) {
      const _filepath = isIndex ? pathe.join(filepath, `index${extension}`) : pathe.join(`${extension}${extension}`);
      if (fs.existsSync(_filepath)) {
        resolvedPath = _filepath;
        break;
      }
    }
    return resolvedPath;
  };
}
function _getSwcOptions(id, tsConfig, options) {
  const isTsFile = /\.tsx?$/.test(id);
  const compilerOptions = tsConfig?.compilerOptions;
  const moduleTarget = compilerOptions?.target?.toString().toLowerCase();
  const defaultOptions = {
    filename: id,
    sourceMaps: compilerOptions?.sourceMap,
    jsc: {
      keepClassNames: compilerOptions?.experimentalDecorators,
      target: moduleTarget,
      parser: {
        syntax: isTsFile ? "typescript" : "ecmascript",
        tsx: isTsFile,
        jsx: !isTsFile,
        decorators: compilerOptions?.experimentalDecorators
      },
      transform: {
        legacyDecorator: compilerOptions?.experimentalDecorators,
        decoratorMetadata: compilerOptions?.emitDecoratorMetadata,
        react: {
          pragma: compilerOptions?.jsxFactory,
          pragmaFrag: compilerOptions?.jsxFragmentFactory,
          importSource: compilerOptions?.jsxImportSource
        }
      }
    }
  };
  return merge__default(defaultOptions, options);
}

function typescript(options, tsConfig) {
  const typescriptOptions = _getTypescriptOptions(options, tsConfig);
  return RollupTypescript__default(typescriptOptions);
}
function _getTypescriptOptions(options, tsConfig) {
  const typescriptOptions = options.rollupPlugins?.typescript || {};
  const compilerOptions = tsConfig?.compilerOptions;
  const defaultOptions = {
    tsconfig: "tsconfig.json",
    tsconfigOverride: { compilerOptions, declaration: false },
    useTsconfigDeclarationDir: true
  };
  return merge__default(defaultOptions, typescriptOptions);
}

async function getRollupPlugins(options, watch) {
  const defaultPlugins = await getDefaultPlugins(options, watch);
  return typeof options.mapRollupPlugins === "function" ? options.mapRollupPlugins(defaultPlugins) : defaultPlugins;
}
async function getDefaultPlugins(options, watch) {
  const tsConfigJson = await loadTsConfig(options.rootDir, options.tsConfig);
  const syntaxParser = resolveSyntaxParser(options, tsConfigJson);
  const runEnabled = watch && options.run;
  return [
    replace__default({
      true: true,
      ...options.rollupPlugins?.replace
    }),
    alias__default({
      ...options.rollupPlugins?.alias
    }),
    resolve__default({
      mainFields: ["module", "jsnext", "main"],
      browser: options.target === "browser",
      extensions: options.extensions,
      exportConditions: [options.target === "node" ? "node" : "browser"],
      preferBuiltins: options.target === "node",
      ...options.rollupPlugins?.resolve
    }),
    json__default({
      preferConst: true,
      ...options.rollupPlugins?.json
    }),
    syntaxParser(),
    commonjs__default({
      include: [/node_modules/],
      extensions: options.extensions,
      ignoreTryCatch: true,
      requireReturnsDefault: "namespace",
      ...options.rollupPlugins?.commonjs
    }),
    raw({
      include: [/\.(md|txt)$/],
      exclude: [],
      ...options.rollupPlugins?.raw
    }),
    runEnabled && run__default({
      ...options.rollupPlugins?.run
    })
  ];
}
function resolveSyntaxParser(options, tsConfigJson) {
  const factory = _resolveParserFactory(options.parser);
  return () => factory(options, tsConfigJson);
}
function _resolveParserFactory(parser) {
  switch (parser) {
    case "swc":
      return swc;
    case "typescript":
      return typescript;
    default:
      return esbuild;
  }
}

async function getRollupOptions(options, watch) {
  const rollupPlugins = await getRollupPlugins(options, watch);
  const defaultOptions = {
    context: options.rootDir,
    input: options.entries,
    output: getRollupOutputs(options),
    plugins: rollupPlugins,
    watch: watch && _getWatchOptions(),
    external: _createExternalFilter(options),
    onwarn: _createWarnHandler()
  };
  return merge__default(defaultOptions, options.rollupOptions);
}
function _getWatchOptions() {
  return {
    chokidar: {
      ignoreInitial: true,
      ignorePermissionErrors: true,
      ignored: ["**/.git/**", "**/node_modules/**"]
    }
  };
}
function _getPackageName(id) {
  const partials = id.split("/");
  return partials[0][0] === "@" ? `${partials[0]}/${partials[1]}` : partials[0];
}
function _createExternalFilter(options) {
  return (id) => {
    const packageName = _getPackageName(id);
    const isExternal = options.externals.includes(packageName);
    if (isExternal) {
      return true;
    }
    if (options.inlineDependencies || id.startsWith(".") || id.startsWith("/") || id.match(/src[\\/]/)) {
      return false;
    }
    return isExternal;
  };
}
function _createWarnHandler() {
  const warningIgnoreList = ["CIRCULAR_DEPENDENCY", "EMPTY_BUNDLE", "THIS_IS_UNDEFINED"];
  return (warning, handler) => {
    if (warning.code === "UNRESOLVED_IMPORT") {
      const id = warning.source;
      const importer = warning.importer;
      if (!importer || !/\?commonjs-external$/.test(importer)) {
        console.error(`Rollup failed to resolve import "${id}" from "${importer}"`);
        process.exit(-1);
      }
    }
    if (!warningIgnoreList.includes(warning.code)) {
      if (warning.code === "PLUGIN_WARNING") {
        console.warn(`[plugin:${warning.plugin}] ${warning.message}`);
      } else {
        handler(warning);
      }
    }
  };
}

async function build(options) {
  const rollupOptions = await getRollupOptions(options);
  await remove(options.outDir);
  return createBuilder(options, rollupOptions);
}

async function watch(options) {
  const rollupOptions = await getRollupOptions(options, true);
  await remove(options.outDir);
  return createWatcher(options, rollupOptions);
}

const configNames = /* @__PURE__ */ new Set([
  ".bauenrc.js",
  ".bauenrc.cjs",
  ".bauenrc.mjs",
  ".bauenrc.ts",
  "bauen.config.js",
  "bauen.config.cjs",
  "bauen.config.mjs",
  "bauen.config.ts"
]);
function searchConfig(rootDir) {
  const result = { name: null, path: null };
  for (const name of configNames) {
    const path = pathe.resolve(rootDir, name);
    if (fs.existsSync(path)) {
      Object.assign(result, { name, path });
      break;
    }
  }
  return result;
}

function loadConfig(rootDir) {
  const _require = jiti__default(rootDir, { interopDefault: true });
  const foundConfig = searchConfig(rootDir);
  try {
    return _require(foundConfig.path);
  } catch (err) {
    console.error(`Config file not found in ${rootDir}`, err);
    process.exit(1);
  }
}

function getCliOptions(argv) {
  const args = mri__default([...argv].splice(2));
  const { _, ...restArgs } = args;
  return {
    ..._.length > 0 && { command: _[0] },
    ...restArgs
  };
}

function getDefaultOptions() {
  return {
    command: "build",
    rootDir: process.cwd(),
    outDir: "dist",
    entries: [],
    outputs: [],
    target: "node",
    extensions: [".ts", ".tsx", ".mjs", ".cjs", ".js", ".jsx"],
    externals: [],
    tsConfig: "tsconfig.json",
    parser: "esbuild",
    run: false,
    declaration: true,
    inlineDependencies: false
  };
}

async function getOptions(argv, userConfig) {
  const defaultOptions = getDefaultOptions();
  const inputOptions = { ...getCliOptions(argv), ...userConfig };
  const rootDir = pathe.resolve(inputOptions.rootDir || defaultOptions.rootDir);
  const tsConfig = pathe.resolve(rootDir, inputOptions.tsConfig || defaultOptions.tsConfig);
  const [packageJson] = await Promise.all([
    loadPackage(rootDir),
    loadTsConfig(rootDir, tsConfig)
  ]);
  return merge__default(defaultOptions, inputOptions, {
    rootDir,
    outDir: pathe.resolve(rootDir, inputOptions.outDir || defaultOptions.outDir),
    entries: inputOptions.entries?.map((entry) => pathe.resolve(rootDir, entry)),
    externals: _getExternals(userConfig, packageJson)
  });
}
function _getExternals(userConfig, packageJson) {
  if (userConfig.target === "browser") {
    return [];
  }
  return [
    ...Module__default.builtinModules,
    ...Module__default.builtinModules.map((mod) => `node:${mod}`),
    ...Object.keys(packageJson.dependencies || []),
    ...Object.keys(packageJson.peerDependencies || []),
    ...userConfig.externals || []
  ];
}

exports.build = build;
exports.createBuilder = createBuilder;
exports.createWatcher = createWatcher;
exports.defineOutput = defineOutput;
exports.esbuild = esbuild;
exports.getCliOptions = getCliOptions;
exports.getDefaultOptions = getDefaultOptions;
exports.getOptions = getOptions;
exports.getRollupOptions = getRollupOptions;
exports.getRollupOutput = getRollupOutput;
exports.getRollupOutputs = getRollupOutputs;
exports.getRollupPlugins = getRollupPlugins;
exports.loadConfig = loadConfig;
exports.loadPackage = loadPackage;
exports.loadTsConfig = loadTsConfig;
exports.raw = raw;
exports.registerOutput = registerOutput;
exports.remove = remove;
exports.searchConfig = searchConfig;
exports.swc = swc;
exports.typescript = typescript;
exports.watch = watch;
