import * as rollup from 'rollup';
import { OutputOptions, RollupOptions, Plugin, RollupOutput } from 'rollup';
import { RollupAliasOptions } from '@rollup/plugin-alias';
import { RollupCommonJSOptions } from '@rollup/plugin-commonjs';
import { RollupJsonOptions } from '@rollup/plugin-json';
import { RollupNodeResolveOptions } from '@rollup/plugin-node-resolve';
import { RollupReplaceOptions } from '@rollup/plugin-replace';
import { RollupRunOptions } from '@rollup/plugin-run';
import { FilterPattern } from '@rollup/pluginutils';
import { Options } from '@swc/core';
import { Options as Options$1 } from 'rollup-plugin-esbuild';
import { RPT2Options } from 'rollup-plugin-typescript2';
import { CompilerOptions } from 'typescript';
import { PathLike } from 'fs';

declare type DeepPartial<T> = {
    [P in keyof T]?: DeepPartial<T[P]>;
};
declare type UserConfig = DeepPartial<Config>;
declare type OutputType = "js" | "cjs" | "esm";
declare type OutputTarget = "node" | "browser";
declare type SyntaxParser = "swc" | "esbuild" | "typescript";
interface RollupSwcOptions extends Options {
    extensions?: string[];
}
interface RollupRawOptions {
    include?: FilterPattern;
    exclude?: FilterPattern;
}
declare type RollupTypescriptOptions = RPT2Options;
interface Config {
    rootDir: string;
    outDir: string;
    entries: string[];
    outputs: Array<OutputType | OutputOptions>;
    target: OutputTarget;
    extensions: string[];
    externals: string[];
    tsConfig: string;
    parser?: SyntaxParser;
    declaration?: boolean;
    run?: boolean;
    preserveModules?: boolean;
    inlineDependencies?: boolean;
    onBundleStart?: () => Promise<void>;
    onBundleEnd?: () => Promise<void>;
    rollupOptions?: RollupOptions;
    rollupPlugins?: PluginsConfig;
    mapRollupPlugins?: (plugins: Plugin[]) => Plugin[];
}
interface PluginsConfig {
    replace?: RollupReplaceOptions;
    alias?: RollupAliasOptions;
    resolve?: RollupNodeResolveOptions;
    json?: RollupJsonOptions;
    swc?: RollupSwcOptions;
    raw?: RollupRawOptions;
    run?: RollupRunOptions;
    esbuild?: Options$1;
    commonjs?: RollupCommonJSOptions;
    typescript?: RollupTypescriptOptions;
}

declare type BauenOptions = CliOptions & Config;
interface CliOptions {
    command?: "build" | "watch";
}

declare type PackageJsonType = "commonjs" | "module";
interface PackageJsonBinTable {
    [name: string]: string;
}
interface PackageJsonDependencyTable {
    [name: string]: string;
}
interface PackageJsonScriptTable {
    [name: string]: string;
}
interface PackageJson {
    name: string;
    version?: string;
    private?: boolean;
    description?: string;
    main?: string;
    module?: string;
    types?: string;
    type?: PackageJsonType;
    bin?: PackageJsonBinTable | string;
    workspaces?: string[];
    scripts?: PackageJsonScriptTable;
    dependencies?: PackageJsonDependencyTable;
    optionalDependencies?: PackageJsonDependencyTable;
    devDependencies?: PackageJsonDependencyTable;
    peerDependencies?: PackageJsonDependencyTable;
}

interface TsConfig {
    extends?: string;
    include?: string[];
    exclude?: string[];
    compilerOptions?: CompilerOptions;
    [option: string]: any;
}

declare function createBuilder(options: BauenOptions, rollupOptions: RollupOptions): Promise<RollupOutput[]>;

declare function createWatcher(options: BauenOptions, rollupOptions: RollupOptions): rollup.RollupWatcher;

declare function getRollupOptions(options: BauenOptions, watch?: boolean): Promise<RollupOptions>;

declare function getRollupOutput(id: string, preserveModules?: boolean): OutputOptions & {
    preserveModules: boolean;
};
declare function getRollupOutputs(options: BauenOptions): OutputOptions[];
declare function registerOutput(id: string, output: OutputOptions, force?: boolean): void;

declare function getRollupPlugins(options: BauenOptions, watch?: boolean): Promise<Plugin[]>;

declare function build(options: BauenOptions): Promise<rollup.RollupOutput[]>;

declare function watch(options: BauenOptions): Promise<rollup.RollupWatcher>;

declare function loadConfig(rootDir: string): DeepPartial<Config>;

declare function searchConfig(rootDir: string): {
    name: null;
    path: null;
};

declare function getCliOptions(argv: string[]): CliOptions;

declare function getDefaultOptions(): BauenOptions;

declare function getOptions(argv: string[], userConfig: UserConfig): Promise<BauenOptions>;

declare function esbuild(options: BauenOptions, tsConfig: TsConfig): rollup.Plugin;

declare function raw(options: RollupRawOptions): Plugin;

declare function swc(options: BauenOptions, tsConfig: TsConfig): Plugin;

declare function typescript(options: BauenOptions, tsConfig: TsConfig): rollup.Plugin;

declare const defineConfig: (config: UserConfig) => DeepPartial<Config>;

declare const defineOutput: (output: OutputOptions) => OutputOptions;

declare function loadPackage(rootDir: string, force?: boolean): Promise<PackageJson>;

declare function loadTsConfig(rootDir: string, filename: string): Promise<TsConfig>;

declare function mergeConfig(config: UserConfig, ...configs: UserConfig[]): DeepPartial<Config>;

declare function remove(path: PathLike): Promise<void>;

export { BauenOptions, CliOptions, Config, DeepPartial, OutputTarget, OutputType, PackageJson, PackageJsonBinTable, PackageJsonDependencyTable, PackageJsonScriptTable, PackageJsonType, PluginsConfig, RollupRawOptions, RollupSwcOptions, RollupTypescriptOptions, SyntaxParser, TsConfig, UserConfig, build, createBuilder, createWatcher, defineConfig, defineOutput, esbuild, getCliOptions, getDefaultOptions, getOptions, getRollupOptions, getRollupOutput, getRollupOutputs, getRollupPlugins, loadConfig, loadPackage, loadTsConfig, mergeConfig, raw, registerOutput, remove, searchConfig, swc, typescript, watch };
